def test_asgardpyconfig():
    from asgardpy.config import AsgardpyConfig

    assert AsgardpyConfig()


def test_asgardpyanalysis():
    from asgardpy.analysis import AsgardpyAnalysis

    assert AsgardpyAnalysis
